#!/usr/bin/env python

# Copyright (C) 2015, Wazuh Inc.
# Created by Wazuh, Inc. <info@wazuh.com>.
# This program is free software; you can redistribute it and/or modify it under the terms of GPLv2

import argparse
import asyncio
import itertools
import logging
import operator
import sys
from datetime import datetime

import wazuh.core.cluster.cluster
import wazuh.core.cluster.utils
from wazuh.core.cluster import control, local_client
from wazuh.core.common import decimals_date_format


def __print_table(data, headers, show_header=False):
    """
    Pretty print list of lists
    """

    def get_max_size_cols(l):
        """
        For each column of the table, return the size of the biggest element
        """
        return list(map(lambda x: max(map(lambda y: len(y) + 2, x)), map(list, zip(*l))))

    if show_header:
        table = list(itertools.chain([tuple(map(lambda x: x.upper(), headers))], data))
    else:
        table = data

    sizes = get_max_size_cols(table)

    table_str = '\n'.join(
        [''.join(["{}{}".format(col, " " * (max_size - len(col))) for col, max_size in zip(row, sizes)])
         for row in table])
    print(table_str)


async def print_agents(filter_status, filter_node):
    lc = local_client.LocalClient()
    result = await control.get_agents(lc, filter_node=filter_node, filter_status=filter_status)
    headers = {'id': 'ID', 'name': 'Name', 'ip': 'IP', 'status': 'Status', 'version': 'Version',
               'node_name': 'Node name'}
    data = map(operator.itemgetter(*headers.keys()), result['items'])
    __print_table(data, list(headers.values()), True)


async def print_nodes(filter_node):
    lc = local_client.LocalClient()
    result = await control.get_nodes(lc, filter_node=filter_node)
    headers = ["Name", "Type", "Version", "Address"]
    data = map(lambda x: list(x.values()), result['items'])
    __print_table(data, headers, True)


async def print_health(config, more, filter_node):
    """Print the current status of the cluster as well as additional information.

    Parameters
    ----------
    config : dict
        Cluster current configuration.
    more : bool
        Indicate whether additional information is desired or not.
    filter_node : str, list
        Node to return.
    """

    def calculate_seconds(start_time, end_time):
        """Calculate the time difference between two dates.

        Parameters
        ----------
        start_time : str
            Start date.
        end_time : str
            End date.

        Returns
        -------
        str
            Total seconds between the two dates.
        """
        if end_time != 'n/a' and start_time != 'n/a':
            seconds = \
                datetime.strptime(end_time, decimals_date_format) - datetime.strptime(start_time, decimals_date_format)
            total_seconds = f"{round(seconds.total_seconds(), 3) if seconds.total_seconds() >= 0.0005 else 0.001}s"
        else:
            total_seconds = 'n/a'

        return total_seconds

    lc = local_client.LocalClient()
    result = await control.get_health(lc, filter_node=filter_node)
    msg2 = ""

    msg1 = f"Cluster name: {config['name']}\n\n"
    msg1 += f"Last completed synchronization for connected nodes ({result['n_connected_nodes']}):\n" if not more \
        else f"Connected nodes ({result['n_connected_nodes']}):"

    for node, node_info in sorted(result["nodes"].items()):
        msg2 += f"\n    {node} ({node_info['info']['ip']})\n"
        msg2 += f"        Version: {node_info['info']['version']}\n"
        msg2 += f"        Type: {node_info['info']['type']}\n"
        msg2 += f"        Active agents: {node_info['info']['n_active_agents']}\n"

        if node_info['info']['type'] != "master":
            if not more:
                msg1 += f"    {node} ({node_info['info']['ip']}): " \
                        f"Integrity check: {node_info['status']['last_check_integrity']['date_end_master']} | " \
                        f"Integrity sync: {node_info['status']['last_sync_integrity']['date_end_master']} | " \
                        f"Agents-info: {node_info['status']['last_sync_agentinfo']['date_end_master']} | " \
                        f"Last keep alive: {node_info['status']['last_keep_alive']}.\n"

            msg2 += "        Status:\n"

            # Last Keep Alive
            msg2 += "            Last keep Alive:\n"
            msg2 += f"                Last received: {node_info['status']['last_keep_alive']}.\n"

            # Integrity check
            total = calculate_seconds(node_info['status']['last_check_integrity']['date_start_master'],
                                      node_info['status']['last_check_integrity']['date_end_master'])
            msg2 += f"            Integrity check:\n"
            msg2 += f"                Last integrity check: {total} " \
                    f"({node_info['status']['last_check_integrity']['date_start_master']} - " \
                    f"{node_info['status']['last_check_integrity']['date_end_master']}).\n"
            msg2 += f"                Permission to check integrity: {node_info['status']['sync_integrity_free']}.\n"

            # Integrity sync
            total = calculate_seconds(node_info['status']['last_sync_integrity']['date_start_master'],
                                      node_info['status']['last_sync_integrity']['date_end_master'])
            msg2 += "            Integrity sync:\n"
            msg2 += f"                Last integrity synchronization: {total} " \
                    f"({node_info['status']['last_sync_integrity']['date_start_master']} - " \
                    f"{node_info['status']['last_sync_integrity']['date_end_master']}).\n"

            n_shared = str(node_info['status']['last_sync_integrity']['total_files']["shared"])
            n_missing = str(node_info['status']['last_sync_integrity']['total_files']["missing"])
            n_extra = str(node_info['status']['last_sync_integrity']['total_files']["extra"])
            n_extra_valid = str(node_info['status']['last_sync_integrity']['total_files']["extra_valid"])
            msg2 += f"                Synchronized files: Shared: {n_shared} | Missing: {n_missing} | " \
                    f"Extra: {n_extra} | Extra valid: {n_extra_valid}.\n"

            msg2 += f"                Extra valid files correctly updated in master: " \
                    f"{node_info['status']['last_sync_integrity']['total_extra_valid']}.\n"

            # Agent info
            total = calculate_seconds(node_info['status']['last_sync_agentinfo']['date_start_master'],
                                      node_info['status']['last_sync_agentinfo']['date_end_master'])
            msg2 += "            Agents-info:\n"
            msg2 += f"                Last synchronization: {total} " \
                    f"({node_info['status']['last_sync_agentinfo']['date_start_master']} - " \
                    f"{node_info['status']['last_sync_agentinfo']['date_end_master']}).\n"
            msg2 += f"                Number of synchronized chunks: " \
                    f"{node_info['status']['last_sync_agentinfo']['n_synced_chunks']}.\n"
            msg2 += f"                Permission to synchronize agent-info: " \
                    f"{node_info['status']['sync_agent_info_free']}.\n"

    print(msg1)
    more and print(msg2)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', action='store_true', dest='debug', help="Enable debug mode")
    parser.add_argument('-fn', '--filter-node', dest='filter_node', nargs='*', type=str, help="Filter by node name")
    parser.add_argument('-fs', '--filter-agent-status', dest='filter_status', nargs='*', type=str,
                        help="Filter by agent status")
    exclusive = parser.add_mutually_exclusive_group()
    exclusive.add_argument('-a', '--list-agents', action='store_const', const='list_agents', help='List agents')
    exclusive.add_argument('-l', '--list-nodes', action='store_const', const='list_nodes', help='List nodes')
    exclusive.add_argument('-i', '--health', action='store', nargs='?', const='health', help='Show cluster health')
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.ERROR, format='%(levelname)s: %(message)s')

    cluster_status = wazuh.core.cluster.utils.get_cluster_status()
    if cluster_status['enabled'] == 'no' or cluster_status['running'] == 'no':
        logging.error("Cluster is not running.")
        sys.exit(1)

    cluster_config = wazuh.core.cluster.utils.read_config()
    wazuh.core.cluster.cluster.check_cluster_config(config=cluster_config)

    try:
        if args.filter_status and not args.list_agents:
            logging.error("Wrong arguments.")
            parser.print_help()
            sys.exit(1)
        elif args.list_agents:
            my_function, my_args = print_agents, (args.filter_status, args.filter_node,)
        elif args.list_nodes:
            my_function, my_args = print_nodes, (args.filter_node,)
        elif args.health:
            more = args.health.lower() == 'more'
            my_function, my_args = print_health, (cluster_config, more, args.filter_node,)
        else:
            parser.print_help()
            sys.exit(0)

        asyncio.run(my_function(*my_args))
    except KeyboardInterrupt:
        pass
    except Exception as e:
        logging.error(e)
        if args.debug:
            raise
